// CriticalSection.h: interface for the CCriticalSection class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_CRITICALSECTION_H__8891B66B_6E81_4637_8FE1_9FC6B865984C__INCLUDED_)
#define AFX_CRITICALSECTION_H__8891B66B_6E81_4637_8FE1_9FC6B865984C__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000



/*
 Description

	Represents a Critical Section:
	allows Locking/ Unlocking of a resource

	Use:
	- Lock()
	  to lock the resource
	- Unlock()
	  to unlock the resource

	For automatic Locking/ Unlocking of a resource, see
	CAutoLockUnlockResource class.

 IMPORTANT

	allows transfer of ownership (just like std::auto_ptr).

	So, when you copy-construct a Critical Section,
	make sure the copy-constructed copy OUTLIVES the original object,
	or that you don't use the original object after the 
	copy-constructed object is destroyed.

	The copy-constructor was provided only so that you can 
	construct temporary objects.

*/
class CCriticalSection  
{
	// you are not allowed assign another object to this
	// we are talking about Critical Sections here
	CCriticalSection & operator = ( const CCriticalSection & Not_Implemented);

public:
	CCriticalSection()
	{
		InitializeCriticalSection( GetCriticalSectionPtr() );
	}

	// copy- constructor
	//
	// note: the primitive Critical Section itself has already been constructed
	CCriticalSection( const CCriticalSection & From)
		: m_bIsOwner( true),
		  m_CriticalSection( From.m_CriticalSection)
	{
		From.m_bIsOwner = false;
	}

	~CCriticalSection()
	{
		if ( m_bIsOwner)
		{
			DeleteCriticalSection( GetCriticalSectionPtr() );
		}
	}

	void Lock()
	{
		EnterCriticalSection( GetCriticalSectionPtr());
	}

	void Unlock()
	{
		LeaveCriticalSection( GetCriticalSectionPtr());
	}



	operator LPCRITICAL_SECTION() const
	{
		return GetCriticalSectionPtr();
	}

private:
	LPCRITICAL_SECTION GetCriticalSectionPtr() const
	{
		return & m_CriticalSection;
	}

private:
	// the critical section itself
	mutable CRITICAL_SECTION m_CriticalSection;


	// true if it owns the Critical Section
	mutable bool m_bIsOwner;
};



/*
 Description
	
	allows automatic Locking/ Unlocking of a Resource,
	protected by a Critical Section:
	- locking when this object gets constructed
	- unlocking when this object is destructed
	  (goes out of scope)

 Usage

	when you need protected access to a resource, do the following
	1. Create a Critical Section associated with the resource
	2. Embody the code accessing the resource in braces {}
	3. Construct an CAutoLockUnlockResource object

  Example:
	// we assume that m_CriticalSection
	// is a private variable, and we use it to protect
	// 'this' from being accessed while we need safe access to a resource
	

	// code that does not need resource locking
	
	{
		CAutoLockUnlockResource I_am_locked( m_CriticalSection);
		
		// code that needs thread locking
	}
	
	// code that does not need resource locking

 Usage

 For more details: see CCriticalSection described above.
*/
class CAutoLockUnlockResource
{
	// not implemented - see Usage!!!
	CAutoLockUnlockResource operator=( CAutoLockUnlockResource & Not_Implemented);
	CAutoLockUnlockResource( const CAutoLockUnlockResource & Not_Implemented);

public:
	CAutoLockUnlockResource( CCriticalSection & csResource)
		: m_csResource( csResource)
	{
		m_csResource.Lock();
	}

	~CAutoLockUnlockResource()
	{
		m_csResource.Unlock();
	}

private:

	// the Critical Section for this resource
	CCriticalSection & m_csResource;
};


#endif // !defined(AFX_CRITICALSECTION_H__8891B66B_6E81_4637_8FE1_9FC6B865984C__INCLUDED_)
