// Event.h: interface for the CEvent class.
//
//////////////////////////////////////////////////////////////////////


/*
	Changes to this file

	Date		Author	Remarks
	6/1/2001	JT		allowed named Events
*/
#if !defined(AFX_EVENT_H__5F7DFE51_AA6A_4014_A0CE_838F19E9449F__INCLUDED_)
#define AFX_EVENT_H__5F7DFE51_AA6A_4014_A0CE_838F19E9449F__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "DebugUtility.h"

/*
	note:

	we embody the CEvent class into the App namespace,
	since MFC has a similarly called class: CEvent.

	If we will use MFC, our class might have collided with the 
	one from MFC
*/
namespace App
{

/*
	the possible results of a Wait 
	(for an Event/ Mutex/ etc.)
*/
typedef enum WaitResultType
{
	// the owner thread (the one that created the event),
	// ended without closing the event
	wrtWaitAbandoned = WAIT_ABANDONED,

	// the Event has been signaled
	wrtSignaled = WAIT_OBJECT_0,

	// timeout - we waited for the event, but
	// it did not occur in the specified interval
	wrtWaitTimeout = WAIT_TIMEOUT,
};


/*

	the possible types of Events
*/
typedef enum EventType 
{
	// when the event becomes signaled,
	// one of the waiting threads is released (starts working again),
	// then the event is set back to non-signaled
	//
	// note that when using this type, you DON'T have to call ResetEvent() 
	// at all; see documentation for more details!!!
	etAutomatic,


	// (default)
	// when the event becomes signaled,
	// all threads wake up
	etManual,
};



/*
 Description

	represents an Event.
	This is an object that can be Signaled/ Not Signaled.

	You can also wait for an Event to become Signaled.
*/
class CEvent  
{
	// don't allow operator =
	CEvent & operator = ( const CEvent & Not_Implemented);

public:


	CEvent( const TCHAR * strEventName = NULL, EventType eEventType = etManual)
		: m_pnRefCount( new int( 1))
	{
		BOOL bType;
		switch ( eEventType)
		{
		case etAutomatic:
			bType = FALSE;
			break;

		case etManual:
			bType = TRUE;
			break;

		default:
			// unknown type of Event
			ASSERT( FALSE);
		};

		// create a Non Signaled event
		m_hEvent = CreateEvent( NULL, bType, FALSE, strEventName);
	}
	
	CEvent( const CEvent & From)
		: m_hEvent( From.m_hEvent),
		  m_pnRefCount( From.m_pnRefCount)
	{
		( *m_pnRefCount)++;
	}

	~CEvent()
	{
		( *m_pnRefCount)--;
		if ( *m_pnRefCount == 0)
		{
			// last reference destroyed
			delete m_pnRefCount;
			CloseHandle( m_hEvent);
		}
	}

	// sets the event to SIGNALED state
	void SetEvent()
	{
		VERIFY( ::SetEvent( m_hEvent) );
	}

	// sets the event to NON-SIGNALED state
	void ResetEvent()
	{
		VERIFY( ::ResetEvent( m_hEvent) );
	}

	// set it to SIGNALED state, 
	// make the threads waiting for it wake up,
	// then set it to NON-SIGNALED state again
	void PulseEvent()
	{
		VERIFY( ::PulseEvent( m_hEvent) );
	}

	// waits until this event becomes signaled
	WaitResultType WaitForMe( DWORD dwMilliseconds = INFINITE) const
	{
		return ( WaitResultType)WaitForSingleObject( m_hEvent, dwMilliseconds);
	}

	// allow convertion to Handle
	operator HANDLE() const
	{
		return m_hEvent;
	}

private:
	// the event handle - to call the primitive functions
	HANDLE m_hEvent;

	// the number of references to the Event
	// (the last one will close it)
	int * m_pnRefCount;
};

};

#endif // !defined(AFX_EVENT_H__5F7DFE51_AA6A_4014_A0CE_838F19E9449F__INCLUDED_)
