// MessageHandlerStream.h: interface for the CMessageHandlerStream class.
//
//////////////////////////////////////////////////////////////////////

#if !defined( MESSAGE_HANDLER_STREAM__H)
#define MESSAGE_HANDLER_STREAM__H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <ostream>
#include <sstream>
#include "DebugUtility.h"
#include <memory>

// forward declaration(s)
template< class char_type = char, class CharTraits = std::char_traits< char_type> >
	class CMessageHandlerStream;

/*
 Description

	represents the stream buffer for a Message Handler Stream
	(see below)

 Implementation details

	NOT thread-safe
*/
template< class char_type = char, class CharTraits = std::char_traits< char_type> >
	class CMessageHandler_streambuf
		: public std::basic_streambuf< char_type, CharTraits>
{
private:
	friend CMessageHandlerStream< char, CharTraits>;

	typedef std::basic_streambuf< char_type, CharTraits> streambuf_type;
	typedef CMessageHandlerStream< char_type, CharTraits> ostream_type;
	// base class
	typedef std::basic_streambuf< char_type, CharTraits> base_class;

	enum
	{
		// bad offset - for positioning functions
		_BADOFF = -1,
	};
	

protected:
	////////////////////////////////////////////////////////////////
	// input, not allowed

	virtual int_type pbackfail(int_type = CharTraits::eof())
	{
		// only for output, not for input
		ASSERT( false);
		return (CharTraits::eof()); 
	}
	
	virtual int showmanyc()
	{
		// only for output, not for input
		ASSERT( false);
		return 0; 
	}
	
	virtual int_type underflow()
	{
		// only for output, not for input
		ASSERT( false);
		return (CharTraits::eof()); 
	}
	
	virtual int_type uflow()
	{
		// only for output, not for input
		ASSERT( false);
		return (CharTraits::eof()); 
	}
	
	virtual std::streamsize xsgetn(char_type * _S, std::streamsize _N)
	{
		// only for output, not for input
		ASSERT( false);
		return 0; 

	}
	

	/////////////////////////////////////////////////////////////////
	// positioning, not allowed (see below why)

	virtual pos_type seekoff(off_type, std::ios_base::seekdir,
		std::ios_base::openmode = std::ios_base::in | std::ios_base::out)
	{
		// we don't allow positioning
		ASSERT( false);
		return (std::streampos( _BADOFF)); 
	}
	
	virtual pos_type seekpos(pos_type,
		std::ios_base::openmode = std::ios_base::in | std::ios_base::out)
	{
		// we don't allow positioning
		ASSERT( false);
		return (std::streampos( _BADOFF)); 
	}
	
	/////////////////////////////////////////////////////////////
	// output functions


	// called to write out from the internal
	// buffer, into the external buffer
	virtual int sync()
	{
		m_pOwnerStream->OnNewMessage( GetStreamBuffer().str() );
		m_pStreamBuffer =
			std::auto_ptr< StringStream>( new StringStream);
		return 0; 
	}

	virtual streambuf_type *setbuf( char_type * buffer, std::streamsize n)
	{
		// note: this function MUST be called
		// before working with this stream buffer


		// we don't use a buffer - we forward everything
		ASSERT( buffer == NULL && n == 0);
		setp( NULL, NULL);

		return this; 
	}


	// write the characters from the buffer 
	// to their real destination
	virtual int_type overflow(int_type nChar = CharTraits::eof())
	{
		if ( CharTraits::not_eof( nChar))
		{
			GetStreamBuffer() << ( char_type)nChar;
		}
		return CharTraits::not_eof( nChar);
	}
 
	
	virtual std::streamsize xsputn(const char_type *S, std::streamsize N)
	{
		GetStreamBuffer().write( S, N);
		return N;
	}

public:
	CMessageHandler_streambuf( )
		: m_pStreamBuffer( new StringStream)
	{
	}

private:
	typedef std::basic_ostringstream< char_type> StringStream;
	StringStream & GetStreamBuffer()
	{
		// allocate the buffer first !!!
		ASSERT ( m_pStreamBuffer.get() );
		return *m_pStreamBuffer;
	}

private:
	// holds the Message, until it's flushed
	std::auto_ptr< StringStream> m_pStreamBuffer;

	// the Message Handler Stream - where we write into
	ostream_type * m_pOwnerStream;
};


/*
 Description

	abstract class

	represents a stream that holds a message until it's full,
	and then, calls an Event function - OnNewMessage

	A message is full when the stream buffer is flushed. 
	For instance, when std::endl is written to a stream.
	
	Example:

	streamOut << "Your name is: '" << strUserName << "'" << std::endl;
	// std::endl marks the end of a Message, for instance
	// 'Your name is: 'John'<Enter>'

 Usage

	derive your class from this, and implement the PROTECTED OnNewMessage function

*/
template< class char_type = char, class CharTraits = std::char_traits< char_type> >
	class CMessageHandlerStream 
		: public std::basic_ostream< char_type, CharTraits>
{
	typedef CMessageHandler_streambuf< char_type, CharTraits> handler_streambuf_type;
	friend handler_streambuf_type;

	typedef std::basic_ostream< char_type, CharTraits> base_class;

protected:
	typedef std::basic_string< char_type> string_type;

	CMessageHandlerStream()
		: m_StreamBuf(),
		  base_class( NULL)
	{
		m_StreamBuf.m_pOwnerStream = this;
		init( & m_StreamBuf);
		m_StreamBuf.pubsetbuf( NULL, 0);
	}
	~CMessageHandlerStream()
	{}

protected:
	virtual void OnNewMessage( const string_type & strNewMessage) = 0;

public:
	// our stream buffer
	handler_streambuf_type m_StreamBuf; 
};

#endif // !defined( MESSAGE_HANDLER_STREAM__H)
