// PairedMessagesLog.h
#ifndef PAIRED_MESSAGES_LOG__H
#define PAIRED_MESSAGES_LOG__H


// VC
// STL generates very long names; no need for this
#pragma warning (disable: 4786)

#include "MessageHandlerStream.h"
#include "CriticalSection.h"
#include <iostream>
#include <string>
#include <fstream>
#include <iomanip>
#include <time.h>
#include <vector>
#include <map>
#include "Event.h"


/*
	How do we handle thread-related issues?
*/
typedef enum ThreadingType
{
	// everything is thread-safe
	ttThreadSafe,

	// don't care about thread-safety
	ttNotThreadSafe,
};


/*
 Description

	creates a Locker, depending on the Threading Type.
	If thread safe: a Critical Section;
	If not thread safe: nothing.
*/
template< ThreadingType eThreadingType>
	class ThreadObjectLocker;
template<>
	class ThreadObjectLocker< ttThreadSafe>
		: public CCriticalSection
{};
template<>
	class ThreadObjectLocker< ttNotThreadSafe>
{
public:
	void Lock() 
	{}
	void Unlock()
	{}
};


/*
	represents the settings that affect logging of the Paired Message 

 Settings:

	LogFileName 
		- this is the log file corresponding to the Paired Message Log
	      (where messages are outputted)
	LogAllMessagesFileName
		- this is the log file where ALL Messages are written
		  (in case you construct the Paired Messages Log to allow this)
	RefreshEveryTime
		- at every written message, we refresh our Log
	AutoRefreshSeconds
		- this is the period at which we do a Paired Message Log refresh;
		  that is, we have a thread on which we re-write the 
		  Paired Messages Log file
		  (in case you construct the Paired Messages Log to allow this)
*/
struct PairedMessageLogSettings
{
private:
	/*
		defaults
	*/
	enum
	{
		DEFAULT_AUTO_REFRESH_PERIOD = 10000,	// milliseconds
	};

public:
	PairedMessageLogSettings(
		const std::string & strLogFileName,
		const std::string & strAllMessagesLogFileName,
		int nAutoRefreshMilliseconds = DEFAULT_AUTO_REFRESH_PERIOD)
		: m_strLogFileName( strLogFileName),
		  m_strAllMessagesLogFileName( strAllMessagesLogFileName),
		  m_nAutoRefreshMilliseconds( nAutoRefreshMilliseconds),
		  m_bRefreshEveryTime( false)
	{
		// Important: the Log File Name should be fullly qualified
		// (Ex: use 'C:\\Temp\\log.txt' instead of 'log.txt')
		//
		// this is because we will re-open the log from time to time
		// if the Current Path changes, problems could appear !!!
	}

	PairedMessageLogSettings(
		const std::string & strLogFileName,
		const std::string & strAllMessagesLogFileName,
		bool bEveryTime)
		: m_strLogFileName( strLogFileName),
		  m_strAllMessagesLogFileName( strAllMessagesLogFileName),
		  m_bRefreshEveryTime( bEveryTime),
		  m_nAutoRefreshMilliseconds( 0)
	{
		// Important: the Log File Name should be fullly qualified
		// (Ex: use 'C:\\Temp\\log.txt' instead of 'log.txt')
		//
		// this is because we will re-open the log from time to time
		// if the Current Path changes, problems could appear !!!
	}



	bool HasAllMessagesLogFile() const
	{
		bool bHas = ( !m_strAllMessagesLogFileName.empty());
		return bHas;
	}

	// the name of the Log to output to
	//
	// this is required
	const std::string m_strLogFileName;


	// the name of the file where we write ALL messages
	// (no matter if they are paired/ Non-paired)
	//
	// Empty string means no such file
	const std::string m_strAllMessagesLogFileName;


	//  Auto Refresh - once at a given Period
	//
	// 0 - no auto-refresh
	int m_nAutoRefreshMilliseconds;

	// if true, we Refresh() every time a message is written
	//
	// note that if we have a Refresh() every time, 
	// it does not make any sense to Auto-Refresh 
	// (that is, you should set to no Auto-Refresh)
	//
	bool m_bRefreshEveryTime;
}; 


// indents the text with a number of tabs
template< class Stream>
	void WriteIndent( Stream & streamOut, int nIndentCount)
{
	const char INDENT = '\t';
	typedef typename Stream::char_type char_type;
	for ( int idxIndent = 0; idxIndent < nIndentCount; ++ idxIndent)
	{
		streamOut << ( char_type) INDENT;
	}
}



/*
 Description

	allows having a log where only Un-Paired messages (that is, the
	errors) are written.

	Rationale

		The idea is that in complex applications you end up in logging a lot of
		messages in pairs, like 
		'Start receiving data for client jt2001...', 
		'End receiving data for client jt2001...', etc.

		When something goes wrong - most likely because of a bug - , 
		especially on a multi-threaded application, you want to log 
		your activity, to see what went wrong, where and why.

		Instead of writing a huge log, which will make it difficult to 
		view and understand the possible problems, you can use this log.
		**** The scope of this log is to isolate the problems from the things
		     that went ok. ****

		We find this class very useful when debugging, but you can use it
		in release versions also.

		It can also be used when tracking unprocessed data
		(that is, what data is in the middle of processing).

	Prerequisites
		
		We introduce the concept of a 'Paired Message':
		A [Paired Message] is composed of two messages: one that describes
		the beginning of an action, and one that describes the end of the
		same action. The first message (the one describing the beginning of the
		action) must occur (be written) first. The second message, can be written
		later (of course, when the action ends).

		An [Un-Paired Message] is an unfinished Paired Message:
		that is, the first message occured (was written), but the 
		second message has not occured yet (was not written yet).

		We also allow writing Non-Paired Messages. 
		A [Non-Paired Message] is a message that does not follow 
		the concept of a [Paired Message].
		For instance, it's an informational message like, 
		"Found the word 'HTML' in file 'C:\test.txt'".
		Note: don't confuse it with an Un-Paired Message!!!

		The concept for our log is this:
		Once a full Paired Message (that is, both start of action and
		end of action messages) has been written,
		this means all went ok. So, we erase it from the log.
		*** We only keep the informational messages and 
		    possible problems in the log ***
		That is, the informational messages - Non-Paired Messages,
		and the possible problems - the Un-Paired Messages.

		Therefor, in case an error/ problem occurs, you can consult this log, and
		easily see the isolated error(s)/ problem(s).
	
	What happens to all written full Paired Messages?

		Of course, when you analyze a problem (viewing the log), by seeing only
		the problems you might be missing the context in which
		the problems occured.

		Therefor, we allow having another log where ALL Messages, no matter what,
		will be written. This is called the 'All Messages Log' 
		(see Construction below).

		In other words, internally we can have *** two *** logs.
		One that contains only the informational messages & possible problems
		(that is, the summary), and one contains ALL Messages (that is, 
		the details - the context in which problems have occured)

	Construction

		We recommend reading the documentation for the Settings class
		(PairedMessageLogSettings), before reading further.
		
		Construction( strLogFileName, strAllMessagesLogFileName)
		- constructs the log, having
		  - the given log file name
		  - also, in parallel, all messages will be written to the ALL Messages Log
		  - refreshing the Log is done every time 
		    (see 'Refreshments, my friend', below)

		Construction( CPairedMessagesLog< ...>::Settings CurrentSettings)
		- constructs the log, having the given Settings  

		Note(s):
		- in case you supply an empty All Messages Log File Name, we will
		  not log any message to the All Messages Log.
		- the Log File Name MUST contain the full path to the log file. 
		  This is because the Log File will be reopened many times 
		  (when re-writing it).
		  Therefor, in case the current directory changes, 
		  the Log File Name must point to the same file!!!

	Refreshments, my friend

		The ALL Messages Log does not need refreshing. This is because we only
		APPEND information to the ALL Messages Log.
		
		However, with the Log (the Paired Messages Log), there's a whole 
		different story. This is because, as each Un-Paired Message becomes
		a full Paired Message (that is, the end of the action message has been
		written), it has to be erased from the Log.
		Therefor, this Log has to be re-written - therefor, refreshed.

		Refreshing can be done in two ways (controllable at construction,
		see Settings class):
		- Refresh every time
			- refreshes the log every time a message is written to it.
			  Advantage: 
			  - at any given time, you will have the most up-to-date log.
			  Disadvantage:
			  - It can be inefficient, in case large amounts of data 
			    get written to the log. However, due to the purpose of this log, 
			    we think this would happen very rarely.
		- Auto Refresh at X seconds
			- once at a given time (X seconds), *** on a separate thread ***
			  Advantage:
			  - very efficient
			  Disadvantage
			  - in case your application crashes, you might not have the 
			    most up-to-date log

		Default: Auto Refresh at 10 seconds.

	Usage

		You can write to this log two types of messages:
		- Paired Messages
		- non-Paired Messages.

		IMPORTANT:
		- Each Paired Message is UNIQUELY identified, by a Message ID.
		- writing a (Paired or Non-Paired) Message is done like 
		  writing to an STL stream
		- each message is terminated with 'std::endl'.

		Writing a Paired Message:
		------------------------
		A. Automatically (a Message ID is automatically assigned - next_paired() )
		
		CPairedMessagesLog< ...>::paired_log MessageLog = log.next_paired();
		MessageLog << "Start receiving from [X115] satellite" << std::endl;
		// ... user code
		MessageLog << "End receiving data from [X115] satellite" << std::endl;

		B. Manually (get a Message ID, and then call paired() function)

		// get a Message ID
		CPairedMessagesLog< ..>::message_id_type MessageID = log.nextmessageid();

		log.paired( MessageID) << "Start receiving from [X115] satellite" << std::endl;
		// ... user code
		log.paired( MessageID) << "End receiving data from [X115] satellite" << std::endl;


		Writing a Non-Paired Message:
		----------------------------
		
		log.not_paired() 
			<< "Satellite [X115] has been checked: no malfunctions." << std::endl;
		
		Other details:
		-------------

		The functions that allow writing messages (ex: next_paired(), not_paired(),
		etc.) return an stream (STL-like)to write to.
		That is, you can pass it to functions that accept references to 
		std::basic_stream< char_type>.

		Indentation(s):
		You can also indent some of the messages you write. The last parameter
		to the next_paired(), paired(), not_paired() functions allow passing
		an indent also.
		We encourage you to indent a message when it conceptually belongs to another
		message.
		
		Example:
		- for instance, you parse directories (and log this)
		- within each directory, you parse each file (and log this).
		You can think of each message related to a file as conceptually 
		belonging to the message related to the directory it belongs to.

		This will allow you to view the Log/ All Messages Log files easier.

		A trimmed example is shown below:

		CPairedMessagesLog< ...>::paired_log DirectoryLog = log.next_paired();
		DirectoryLog 
			<< "Start parsing directory" << strCurrentDirectory << std::endl;
			
			// for each file
			CPairedMessagesLog< ...>::paired_log FileLog = log.next_paired( 1);
			FileLog 
				<< "Start parsing file" << strCurrentFile << std::endl;
				
				// for each file

			FileLog 
				<< "End parsing file" << strCurrentFile  << std::endl;

		DirectoryLog 
			<< "End parsing directory" << strCurrentDirectory  << std::endl;

		Note(s):
		- Indenting is done using tabs.



 Template parameters:
	char_type 
		- the character type used to print out messages
	eThreadingType 
		- thread safety.
		  Important:
		  if you plan to use this class from multiple threads, you
		  should set this to Thread Safe.

	To see the settings, see description of PairedMessageLogSettings above
	
*/
template< 
		class char_type = char, ThreadingType eThreadingType = ttThreadSafe >
	class CPairedMessagesLog
{

	enum
	{
		// how many characters does each Message ID take (in characters)? 
		MAX_MESSAGE_ID_CHARS = 5,
	};

public:
	// what do we use to hold a Message ID?
	typedef long message_id_type;

	// the class holding this Log's Settings
	typedef PairedMessageLogSettings Settings;

private:

	typedef std::basic_string< char_type> string_type;
	typedef ThreadObjectLocker< eThreadingType> ObjectLocker;

	typedef CPairedMessagesLog< char_type, eThreadingType> this_class;

	/*
		contains information about a Message to be written to log
	*/
	struct MessageInfo
	{
		MessageInfo( int nIndentCount, const string_type & strMessage)
			: m_nIndentCount( nIndentCount),
			  m_strMessage( strMessage)
		{}

		// with how much to indent the message
		const int m_nIndentCount;

		// the message itself
		const string_type m_strMessage;
	};


	/*
		constains information about an NON-PAIRED Message to be written to log
	*/
	struct NonPairedMessageInfo
	{
		NonPairedMessageInfo( 
			message_id_type nWriteAfter, int nIndentCount, const string_type & strMessage)
			: m_nWriteAfter( nWriteAfter),
			  m_nIndentCount( nIndentCount),
			  m_strMessage( strMessage)
		{}

		// after which Paired message should it be written !!!
		message_id_type m_nWriteAfter;

		// the indenting of the Message
		int m_nIndentCount;

		// the message itself
		string_type m_strMessage;
	};


	/*
		helper class

		allows writing an non-paired message
	*/
	class CNonPairedLog
		: public CMessageHandlerStream< char_type>
	{
	public:
		typedef CPairedMessagesLog< char_type, eThreadingType> paired_message_class;

		CNonPairedLog( paired_message_class * pMessagesLog, int nIndentCount)
			: m_pMessagesLog( pMessagesLog),
			  m_nIndentCount( nIndentCount)
		{}

	protected:
		virtual void OnNewMessage( const string_type & strNewMessage)
		{
			m_pMessagesLog->WriteNonPairedMessage( m_nIndentCount, strNewMessage);
		}

	private:
		paired_message_class * m_pMessagesLog;
		int m_nIndentCount;
	};


	/*
		helper class

		allows writing a Paired Message
	*/
	class COnePairedMessageLog
		: public CMessageHandlerStream< char_type>
	{
	public:
		typedef CPairedMessagesLog< char_type, eThreadingType> paired_message_class;

		COnePairedMessageLog( 
				paired_message_class * pMessagesLog, 
				message_id_type MessageID, 
				int nIndentCount)
			: m_pMessagesLog( pMessagesLog),
			  m_MessageID( MessageID),
			  m_nIndentCount( nIndentCount)
		{}

	protected:
		virtual void OnNewMessage( const string_type & strNewMessage)
		{
			m_pMessagesLog->WritePairedMessage( 
				m_MessageID, m_nIndentCount, strNewMessage);
		}

	private:
		// the Log we actually write to 
		paired_message_class * m_pMessagesLog;
		
		// the ID of the message we're about to write
		message_id_type m_MessageID;

		// with how much to indent the message
		int m_nIndentCount;
	}; // COnePairedMessageLog



public:
	// helper for clients of this class, to be able to hold
	// paired logs
	typedef COnePairedMessageLog paired_log;



	CPairedMessagesLog( 
			const std::string & strLogFileName, 
			const std::string & strAllMessagesLogFileName)
		: m_Settings( strLogFileName, strAllMessagesLogFileName)
	{
		InitSettings();
	}

	CPairedMessagesLog( const Settings & CurrentSettings)
		: m_Settings( CurrentSettings)
	{
		InitSettings();
	}

	~CPairedMessagesLog()
	{
		// last refresh
		Refresh();

		// tell the Auto Refresh thread to end, and wait for it
		m_eventStop.SetEvent();
		while ( !m_bAutoRefreshThreadEnded)
		{
			Sleep( 1);
		}
	}

	// writes an Non-paired message
	//
	// a Non-paired message is a message that appears only once;
	// it is ALWAYS printed to the log
	void WriteNonPairedMessage( int nIndentCount, const string_type & strNonPairedMessage)
	{
		m_csObjectLocker.Lock();
		message_id_type CurrentMessageID = m_CurrentMessageID;
		m_aNonPairedMessages.push_back( 
			NonPairedMessageInfo( CurrentMessageID, nIndentCount, strNonPairedMessage));
		if ( m_Settings.m_bRefreshEveryTime)
		{
			// every non-paired Message is appended 
			AppendMessageToLog( 
				nIndentCount, GetNoSurroundedMessage(), strNonPairedMessage);
		}

		if ( m_Settings.HasAllMessagesLogFile() )
		{
			WriteIndent( m_streamAllMessages, nIndentCount);
			m_streamAllMessages << GetNoSurroundedMessage();
			m_streamAllMessages << strNonPairedMessage;
			m_streamAllMessages.flush();
		}

		m_csObjectLocker.Unlock();
	}


	// writes a Paired Message
	void WritePairedMessage( 
		message_id_type MessageID, 
		int nIndentCount,
		const string_type & strMessage)
	{
		m_csObjectLocker.Lock();

		MessagesCollection::iterator itExists = m_collMessages.find( MessageID);
		if ( itExists != m_collMessages.end() )
		{
			// this is the pair of the first Paired Message
			m_collMessages.erase( MessageID);
			InternalRefresh();
		}
		else
		{
			// this is the first Paired Message - insert it
			MessagesCollection::value_type 
				ValToInsert( MessageID, MessageInfo( nIndentCount, strMessage));
			m_collMessages.insert( ValToInsert);
			if ( m_Settings.m_bRefreshEveryTime)
			{
				// it's a new Paired Message - we can append it
				// (that is, we don't have to erase anything from the Log,
				// in order for it to be up-to-date)
				AppendMessageToLog( 
					nIndentCount, 
					GetSurroundedMessageID( MessageID), 
					strMessage);
			}
		}

		if ( m_Settings.HasAllMessagesLogFile() )
		{
			WriteIndent( m_streamAllMessages, nIndentCount);
			m_streamAllMessages << GetSurroundedMessageID( MessageID);
			m_streamAllMessages << strMessage;
			m_streamAllMessages.flush();
		}
		m_csObjectLocker.Unlock();
	}

	void Refresh()
	{
		m_csObjectLocker.Lock();
		InternalRefresh();
		if ( m_Settings.HasAllMessagesLogFile() )
		{
			m_streamAllMessages.flush();
		}
		m_csObjectLocker.Unlock();
	}

	// returns a reference that can write Not-paired messages
	CNonPairedLog not_paired( int nIndentCount = 0)
	{
		return CNonPairedLog( this, nIndentCount);
	}


	// returns a reference that can write a Paired message
	COnePairedMessageLog paired( message_id_type UniqueID, int nIndentCount = 0)
	{
		return COnePairedMessageLog( this, UniqueID, nIndentCount);
	}

	// returns a reference that can write a Paired message
	// 
	// the UniqueID is computed automatically
	COnePairedMessageLog next_paired( int nIndentCount = 0)
	{
		message_id_type UniqueID = nextmessageid();
		return COnePairedMessageLog( this, UniqueID, nIndentCount);
	}


	
	message_id_type nextmessageid() 
	{
		m_csObjectLocker.Lock();
		message_id_type NextID = ++m_CurrentMessageID;
		m_csObjectLocker.Unlock();
		return NextID;
	}

private:

	// note: this object should be locked when 
	// entering this function !!!
	void AppendMessageToLog( 
		int nIndentCount, 
		const string_type & strMessagePrefix, 
		const string_type & strMessage )
	{
		std::basic_ofstream< char_type> 
			streamPairedLog( 
				m_Settings.m_strLogFileName.c_str(), 
				std::ios_base::out | std::ios_base::app );
		WriteIndent( streamPairedLog, nIndentCount);
		streamPairedLog << strMessagePrefix << strMessage;
	}

	// re-freshes the Log
	//
	// note: this object should be locked when 
	// entering this function !!!
	void InternalRefresh()
	{
		std::basic_ofstream< char_type> 
			streamPairedLog( m_Settings.m_strLogFileName.c_str());

		NonPairedMessagesArray::const_iterator
			itFirstNonPaired = m_aNonPairedMessages.begin(),
			itLastNonPaired = m_aNonPairedMessages.end();
		
		// we have some unpaired messages to write
		MessagesCollection::iterator
			itFirstMessage = m_collMessages.begin(),
			itLastMessage = m_collMessages.end();
		while ( itFirstMessage != itLastMessage)
		{
			const message_id_type nCurrentMessageID = itFirstMessage->first;
			// see if any Non-Paired messages have to 
			// be written first !!!
			while ( itFirstNonPaired != itLastNonPaired)
			{
				const message_id_type nWriteAfter = itFirstNonPaired->m_nWriteAfter;
				if ( nWriteAfter < nCurrentMessageID)
				{
					const NonPairedMessageInfo & Info = *itFirstNonPaired;
					WriteIndent( streamPairedLog, Info.m_nIndentCount);
					streamPairedLog 
						<< GetNoSurroundedMessage()
						<< Info.m_strMessage;
				}
				else
				{
					break;
				}
				++ itFirstNonPaired;
			}

			// write the Paired Message
			MessageInfo & Info = itFirstMessage->second;
			WriteIndent( streamPairedLog, Info.m_nIndentCount);
			streamPairedLog 
				<< GetSurroundedMessageID( nCurrentMessageID)
				<< Info.m_strMessage;
			++ itFirstMessage;
		}

		while ( itFirstNonPaired != itLastNonPaired)
		{
			const NonPairedMessageInfo & Info = *itFirstNonPaired;
			WriteIndent( streamPairedLog, Info.m_nIndentCount);
			streamPairedLog 
				<< GetNoSurroundedMessage()
				<< Info.m_strMessage;
			++ itFirstNonPaired;
		}
	}

	void InitSettings()
	{
		m_CurrentMessageID = 1;
		if ( m_Settings.HasAllMessagesLogFile())
		{
			m_streamAllMessages.open( m_Settings.m_strAllMessagesLogFileName.c_str());
		}

		{
		// clear the Log (if previously written)
		std::basic_ofstream< char_type> 
			streamPairedLog( 
				m_Settings.m_strLogFileName.c_str(), std::ios_base::out);
		}

		if ( m_Settings.m_nAutoRefreshMilliseconds != 0)
		{
			DWORD dwThreadID;
			CreateThread( NULL, 0, &this_class::AutoRefreshThread, this, 0, &dwThreadID);
			m_bAutoRefreshThreadEnded = false;
		}
		else
		{
			// (there is no Auto Refresh thread)
			m_bAutoRefreshThreadEnded = true;
		}
	}

	static DWORD WINAPI AutoRefreshThread( LPVOID lpParam)
	{
		this_class * pThis = ( this_class *)lpParam;

		// we should have a Refresh period, in order to start this thread
		ASSERT( pThis->m_Settings.m_nAutoRefreshMilliseconds > 0);

		bool bContinue = true;
		while ( bContinue)
		{
			App::WaitResultType eWaitResult =
				pThis->m_eventStop.WaitForMe( pThis->m_Settings.m_nAutoRefreshMilliseconds);
			switch ( eWaitResult)
			{
			case App::wrtSignaled:
			case App::wrtWaitAbandoned:
				// object is to be destroyed
				bContinue = false;
				pThis->m_bAutoRefreshThreadEnded = true;
				break;

			case App::wrtWaitTimeout:
				pThis->Refresh();
				break;

			default:
				// unknown type of Refresh
				ASSERT( FALSE);
			}
		}

		return 0;
	}

	string_type GetSurroundedMessageID( message_id_type ID) const
	{
		std::basic_stringstream< char_type> streamOut;
		streamOut 
			<< ( char_type)'[' 
			<< std::setw( MAX_MESSAGE_ID_CHARS) 
			<< ID 
			<< ( char_type)']'
			<< ( char_type)' ';
		string_type strSurrounded = streamOut.str();
		return strSurrounded;
	}

	string_type GetNoSurroundedMessage() const
	{
		string_type strNoSurroundedMessage;
		strNoSurroundedMessage.resize( MAX_MESSAGE_ID_CHARS, ( char_type)'-');
		strNoSurroundedMessage = 
			( char_type)'['  + strNoSurroundedMessage + ( char_type)']' + 
			( char_type)' ';
		return strNoSurroundedMessage;
	}

private:


	// the stream for write ALL messages
	// (this is open in case we want to log All Messages also)
	std::basic_ofstream< char_type> m_streamAllMessages;


	// Logging settings 
	//
	// note that they can only be set at construction
	const Settings m_Settings;


	// allow locking the current object - if (when) necessary
	ObjectLocker m_csObjectLocker;


	// contains the current Unique message ID
	//
	// use nextobjectid() to get a unique message ID you can use
	message_id_type m_CurrentMessageID;


	typedef std::vector< NonPairedMessageInfo> NonPairedMessagesArray;
	// contains all the Non-paired Messages that are to be written to the log
	NonPairedMessagesArray m_aNonPairedMessages;


	typedef std::map< message_id_type, MessageInfo> MessagesCollection;
	// contains the Messages that have not yet found their pair
	//
	// every message, when it finds its pair 
	// is removed from here
	MessagesCollection m_collMessages;


	// true when the Auto Refresh Thread has ended
	// (note that the destructor has to wait until this becomes true)
	volatile bool m_bAutoRefreshThreadEnded;

	// set on destruction of this object
	// (to notify any other running threads belonging to this object)
	App::CEvent m_eventStop;
};



#endif // PAIRED_MESSAGES_LOG__H