// TestPairedMessagesLog.cpp 
//

#include "stdafx.h"
#include <windows.h>
#include "PairedMessagesLog.h"
#include <iostream>

#include <exception>
#include <string>
#include <fstream>

#include <direct.h>
#include <io.h>
#include <errno.h>

//////////////////////////////////////////////////////////////////////////
namespace // anonymous namespace
{

	const char APP_LOG_FILE_NAME [] = 
		"D:\\john\\pairedmessages\\log.txt";
	const char APP_LOG_ALL_FILE_NAME [] = 
		"D:\\john\\pairedmessages\\log_all.txt";

	const char DIRECTORY_DELIMETER = '\\';
	const char DIRECTORY_CURRENT [] = ".";
	const char DIRECTORY_PARENT [] = "..";

	const char ALL_FILES [] = "*.*";

	// max. number of concurrent threads, when operating on a set of files
	const int MAX_CONCURRENT_THREAD_ACTIONS = 10;

	// the number of running threads at a given time
	LONG s_nRunningThreads;


	// the path we operate on, on our test
	std::string s_strFoldersPath;

	// when examining files, the Word to search for
	std::string s_strWordToSearchFor;


	CPairedMessagesLog<> & GetAppLog()
	{
		static CPairedMessagesLog<> s_AppLog( 
			CPairedMessagesLog<>::Settings( APP_LOG_FILE_NAME, APP_LOG_ALL_FILE_NAME,
				//2000));				// auto refresh at X milliseconds
				true));			// refresh each time
		return s_AppLog;
	}

	void SearchFileForWord( const std::string & strFileName, const std::string & strWord)
	{
		const int MAX_LINE_SIZE = 2048;
		char strLine[ MAX_LINE_SIZE];

		std::ifstream streamIn( 
			strFileName.c_str(), std::ios_base::in | std::ios_base::binary );

		// will throw an error in case of any error opening the file
		streamIn.exceptions( std::ios_base::failbit);
		// now we only care about exceptions while reading the file
		streamIn.exceptions( std::ios_base::badbit);

		do
		{
			streamIn.getline( strLine, MAX_LINE_SIZE);
			int nReadCharsCount = streamIn.gcount();
			std::string strCurrentLine;
			strCurrentLine.append( strLine, nReadCharsCount );
			if ( strCurrentLine.find( strWord) != std::string::npos)
			{
				GetAppLog().not_paired() 
					<< "Found " << strWord << " in " << strFileName << std::endl;
				break;
			}
		}
		while ( streamIn.good() );
	}


	/*
		does an operation for each file in a given path

	  Remarks
		
		don't use directly, use 'do_for_each_file' function instead
	*/
	template< class T, int nMaxConcurrentThreads>
		class do_for_each_file_t
	{
	public:
		do_for_each_file_t( const std::string & strFullPath, const char *strFileSpec,  T Function)
			: m_Function( Function),
			  m_strFileSpec( strFileSpec),
			  m_nRunningThreads( 0)
		{
			DoForEach( strFullPath);
		}

		~do_for_each_file_t()
		{
			// wait for all threads to end
			while ( m_nRunningThreads > 0)
			{
				Sleep( 1);
			}
		}

	private:
		void DoForEach( const std::string & strPath)
		{
			// cannot have empty path
			ASSERT( !strPath.empty() );
			// make sure the path does not end in "\\"
			ASSERT( ( *strPath.rbegin()) != DIRECTORY_DELIMETER);

			std::string strLookFor = strPath + DIRECTORY_DELIMETER + m_strFileSpec;
			_finddata_t findInfo;
			long hFound = _findfirst(  strLookFor.c_str(), &findInfo);
			bool bContinue = ( hFound != -1);
			while ( bContinue)
			{
				bool bIsDirectory = (( findInfo.attrib & _A_SUBDIR) != false);
				if ( bIsDirectory)
				{
					bool bIsSubDirectory =
						( ( findInfo.name != std::string( DIRECTORY_CURRENT) ) &&
						  ( findInfo.name != std::string( DIRECTORY_PARENT) ));
					if ( bIsSubDirectory)
					{
						std::string strSubdir = 
							strPath + DIRECTORY_DELIMETER + findInfo.name;
						DoForEach( strSubdir);
					}
				}
				else
				{
					std::string strFileName = 
						strPath + DIRECTORY_DELIMETER + findInfo.name;
					ApplyFunction( strFileName);
				}
				bContinue = ( _findnext( hFound, &findInfo) != -1);
			}
		}

		/*
			helper, in order to allow passing data to the Run Function thread
		*/
		struct ThreadData
		{
			ThreadData( 
					do_for_each_file_t< T, nMaxConcurrentThreads> * pThis,
					const std::string & strFileName)
				: m_pThis( pThis),
				  m_strFileName( strFileName)
			{}
			do_for_each_file_t< T, nMaxConcurrentThreads> * m_pThis;
			
			std::string m_strFileName;
		};

		void ApplyFunction( const std::string & strFileName)
		{
			ThreadData * pData = new ThreadData( this, strFileName);
			// first, wait for an available thread
			while ( m_nRunningThreads >= nMaxConcurrentThreads)
			{
				Sleep( 1);
			}
			InterlockedIncrement( &m_nRunningThreads);
			DWORD dwThreadID;
			CreateThread( 
				NULL, 0, 
				&do_for_each_file_t< T, nMaxConcurrentThreads>::RunFunctionThread, 
				pData, 0, &dwThreadID);
		}

		static DWORD WINAPI RunFunctionThread( LPVOID lpParam)
		{
			ThreadData * pData = ( ThreadData *)lpParam;
			pData->m_pThis->m_Function( pData->m_strFileName);
			/*
			try
			{
				pData->m_pThis->m_Function( pData->m_strFileName);
			}
			catch ( std::exception & exc)
			{
				GetAppLog().not_paired() << "Exception " << exc.what() << std::endl ;
			}
			catch ( ...)
			{
				GetAppLog().not_paired() << "Unknown Exception " << std::endl ;
			}
			*/
			InterlockedDecrement( &( pData->m_pThis->m_nRunningThreads));
			delete pData;
			return 0;
		}

	private:
		// the thing to do for each File
		T m_Function;

		// the file specification to look for (ex: "*.txt")
		std::string m_strFileSpec;

		// the number of running threads at a given time
		LONG m_nRunningThreads;
	};

	template< class T>
		inline void do_for_each_file( const std::string & strPath, T Function)
	{
		do_for_each_file_t< T, MAX_CONCURRENT_THREAD_ACTIONS> 
			do_action( strPath, ALL_FILES, Function);
	}



	/*
	 Description

		searches for a word in a file;
		if the word found, we print this to our log 
	*/
	class FindWordInFile
	{
	public:
		FindWordInFile( const std::string & strWord)
			: m_strWord( strWord)
		{}

		void operator() ( const std::string & strFileName)
		{
			CPairedMessagesLog<>::paired_log Log = GetAppLog().next_paired();
			Log << "Start Examining: " << strFileName << std::endl;

			SearchFileForWord( strFileName, m_strWord);

			Log << "End Examining: " << strFileName << std::endl;
		}

	private:
		// the word to look for, in a file
		std::string m_strWord;
	};

	/*
		allows expiring (deleting) files
	*/
	class ExpireFiles
	{
	public:
		ExpireFiles( int nPeriod)
			: m_nPeriod( nPeriod),
			  m_idxFile( 0)
		{}

		void operator() ( const std::string & strFileName)
		{
			if ( ( m_idxFile % m_nPeriod) == 0)
			{
				DeleteFile( strFileName.c_str());
			}
			++ m_idxFile;
		}

	private:
		// the index of the current file
		int m_idxFile;

		// represents the period at which a file will be deleted
		// (that is, one file from a group of files will be deleted)
		int m_nPeriod;
	};



	// CHG JT, 10/19/2001
	// Description
	// 
	//		examines the Folder, and its subfolders, for a given word
	// 
	// Parameters
	// const char * strFolderPath
	// Return Value
	// 
	// Remarks
	// 
	//		
	DWORD WINAPI ExamineFolderAndSubfoldersThread( LPVOID /* lpParam */)
	{
		InterlockedIncrement( &s_nRunningThreads);
		do_for_each_file( s_strFoldersPath, FindWordInFile( s_strWordToSearchFor) );
		InterlockedDecrement( &s_nRunningThreads);
		return 0;
	}


	// CHG JT, 10/25/2001
	// Description
	// 
	//		runs the Expire Files function
	// 
	// Parameters
	// LPVOID /* lpParam */
	// Return Value
	// 
	// Remarks
	// 
	//		
	DWORD WINAPI ExpireFilesThread( LPVOID /* lpParam */)
	{
		InterlockedIncrement( &s_nRunningThreads);

		CPairedMessagesLog<>::paired_log Log = GetAppLog().next_paired();
		Log << "Expiring files started..." << std::endl;

		int nPeriod = 7;
		// we erase all files - but in groups
		while ( nPeriod > 0)
		{
			do_for_each_file( s_strFoldersPath, ExpireFiles( nPeriod) );
			-- nPeriod;
		}
		Log << "Expiring files ended: " << std::endl;

		InterlockedDecrement( &s_nRunningThreads);
		return 0;
	}

}; // anonymous namespace
//////////////////////////////////////////////////////////////////////////



void TestPairedMessagesLog()
{
	/*
		BEWARE

		Some of the files in the given path below, will be DELETED !!!
		Make sure the path points to some temporary data, or otherwise
		you MIGHT LOSE IMPORTANT INFORMATION !!!

		Then, you can continue with the test.
	*/
	std::string strPath = "D:\\john\\pairedmessages\\test";
	std::string strWordSearchFor = "HTML";
	ASSERT( FALSE);
	s_strFoldersPath = strPath;
	s_strWordToSearchFor = strWordSearchFor;

	// make sure our log gets created, first!
	GetAppLog();

	DWORD dwThreadID;
	CreateThread( NULL, 0, ExamineFolderAndSubfoldersThread, NULL, 0, &dwThreadID);

	// wait a while before expiring
	Sleep( 5000);
	CreateThread( NULL, 0, ExpireFilesThread, NULL, 0, &dwThreadID);

	// wait for the test to end
	while ( s_nRunningThreads > 0)
	{
		Sleep( 10);
	}
}

