#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#include <time.h>

#if defined __alpha
#include <c_asm.h>
#endif

#include "memtrack.h"

#if defined MULTI_THREAD

#include <signal.h>
#include <setjmp.h>
#include <pthread.h>

static void mutexOp(int lock)
{
    static pthread_mutex_t mutex;
    static int mutexInit = 0;
    pthread_mutexattr_t mutexAttr;
    
    if ( ! mutexInit) 
    {
	mutexInit = 
	    pthread_mutexattr_init(&mutexAttr) == 0
	    &&
	    pthread_mutexattr_settype(&mutexAttr, PTHREAD_MUTEX_RECURSIVE) == 0
	    &&
	    pthread_mutex_init(&mutex, &mutexAttr) == 0;
	
	if (! mutexInit)
	{
	    fprintf(stderr, 
		    "%s:%d Cannot pthread_mutex_init\n",
		    __FILE__, __LINE__);
	    exit(1);
	}
    }
    
    if ((lock ? pthread_mutex_lock(&mutex) : pthread_mutex_unlock(&mutex)) != 0)
    {
	fprintf(stderr, 
		"%s:%d Cannot %s\n",
		__FILE__, __LINE__, lock ? "pthread_mutex_lock" : "pthread_mutex_release");
	exit(1);
    }
    return;
}

static jmp_buf jmpenv;
static void segvHandler(int signo)
{
    longjmp(jmpenv, 1);
}

static struct sigaction oaction;

#define SET_SIGSEGV_ACTION { \
        struct sigaction action; \
	action.sa_sigaction = NULL; \
	action.sa_handler = segvHandler; \
	action.sa_mask = 0; \
	action.sa_flags = SA_RESTART; \
	sigaction(SIGSEGV, &action, &oaction); }

#define RESET_SIGSEGV_ACTION sigaction (SIGSEGV, &oaction, NULL);

#define MUTEX_LOCK mutexOp(1)
#define MUTEX_RELEASE mutexOp(0);

#define SETJMP(e) setjmp(e)

#else

#define SET_SIGSEGV_ACTION
#define RESET_SIGSEGV_ACTION
#define MUTEX_LOCK
#define MUTEX_RELEASE
#define SETJMP(e) (0)

#endif

/******************************************************************
 * A Pointer may point to a heap allocation if it is 
 * above the uninitailized data and under the program break point
 */
static int isPtrHeap(void* target)
{
    extern unsigned long _end;

    if (target >= (void*)&_end && target < sbrk(0))
    {
	return 1;
    }
   
    return 0;
}

/******************************************************************
* Memory (heap) tracking functions are intended for use
* during tesing only. 
* To enable them, add the text 
* 	-Dcalloc=MEMTRACK_CALLOC 
*       -Dfree=MEMTRACK_FREE
*       -Dmalloc=MEMTRACK_MALLOC 
* 	-Drealloc=MEMTRACK_REALLOC 
*       -Dvalloc=MEMTRACK_VALLOC 
*       -Dstrdup=MEMTRACK_STRDUP 
*       -Dstrcat=MEMTRACK_STRCAT
*       -Dstrcpy=MEMTRACK_STRCPY
* to the CFLAGS in your makefile
* The functions keep a record of all memory allocated through them
* and alert the user of any questionable behavior.
*/

#undef calloc
void* calloc(size_t numElements, size_t elemSize);
#undef free
void free(void* ptr);
#undef malloc
void* malloc(size_t size);
#undef realloc
void* realloc(void* ptr, size_t size);
#undef valloc
void* valloc(size_t size);
#undef strdup
char* strdup(const char* s1);
#undef strcpy
char* strcpy(char* s1, const char* s2);
#undef strcat
char* strcat(char* s1, const char* s2);

#define NO_REF        ((char)0)
#define REF_TO_START  ((char)1)
#define REF_TO_WITHIN ((char)2)

#define LEAK_REPORTED ((char)1)
#define POTENTIAL_LEAK_REPORTED ((char)2)

#define MAX_FN 13
typedef struct 
{
    void*  mem;
    size_t size;
    ushort fileLine;
    char   fileName[MAX_FN];
    char   refExists;
    char   reported;
} AllocRec;

#define MEMLIST_CHUNK 512

/* Keep the FREED_FLAG a printable character so that 
 * freed strings, when displayed, are clearly seen */
#define FREED_FLAG '*' 
#define INIT_FLAG '!'
#define END_SENTINEL '$'
#define AUTOCHECK_PERIOD 15

static AllocRec* memList = NULL;
static size_t memListConsumed = 0;
static size_t memListAllocated = 0;
static void* lowAddress = (void*)ULONG_MAX;
static void* highAddress = 0;

static void markAsFree(void* ptr, size_t size)
{
    memset(ptr, FREED_FLAG, size);
    return;
}

static AllocRec* memList_realloc(size_t newSize)
{
    AllocRec* newMemList = calloc(newSize, sizeof(AllocRec));
    assert(newMemList != NULL);
    if (memList)
    {
	memcpy((void*)newMemList, (void*)memList, memListConsumed * sizeof(AllocRec));
	markAsFree(memList, memListConsumed * sizeof(AllocRec));
	free(memList);
    }
    return newMemList;
}

static void memList_init(void)
{
    if (memListConsumed == memListAllocated)
    {
	memListAllocated += MEMLIST_CHUNK;
	memList = memList_realloc(memListAllocated);
	assert(memList != NULL);
    }
    else 
    {
	assert (memListConsumed < memListAllocated);
	if (memListAllocated - memListConsumed > 2*MEMLIST_CHUNK)
	{
	    memListAllocated -= MEMLIST_CHUNK;
	    memList = memList_realloc(memListAllocated);
	    assert(memList != NULL);
	}
    }
    return;
}

static void memList_add(volatile void* mem, size_t size, char* fn, ushort fl)
{
    if (mem == NULL) return;

    memList_init();
    memList[memListConsumed].mem = (void*)mem;
    memList[memListConsumed].size = size;
    memList[memListConsumed].refExists = NO_REF;
    memList[memListConsumed].reported = 0;

    if (fn != NULL)
    {
	strncpy(memList[memListConsumed].fileName, fn, MAX_FN); 
	memList[memListConsumed].fileName[MAX_FN-1] = 0;
	memList[memListConsumed].fileLine = fl;
    }
    else
    {
	memList[memListConsumed].fileName[0] = 0;
	memList[memListConsumed].fileLine = 0;
    }

    ++memListConsumed;

    if (mem < lowAddress ) lowAddress = (void*)mem;
    if (mem > highAddress ) highAddress = (void*)mem;

    memtrackCheckPeriodic(AUTOCHECK_PERIOD);

    return;
}

static void findAddressExtremes(void)
{
    int ix;

    lowAddress = (void*)ULONG_MAX;
    highAddress = 0;

    for (ix = 0; ix < memListConsumed; ++ix)
    {
	void* mem = memList[ix].mem;
	if (mem < lowAddress ) lowAddress = (void*)mem;
	if (mem > highAddress ) highAddress = (void*)mem;
    }
	
    return;
}

static AllocRec memList_remove(void* mem)
{
    AllocRec* ptrMemList;
    AllocRec* ptrMemListEnd;
    AllocRec ret;

    memList_init();
    ptrMemListEnd = &memList[memListConsumed];
    memset(&ret, 0, sizeof(ret));
    if (memList != NULL && memListConsumed) 
    {
	for (ptrMemList = memList; 
	     ptrMemList < ptrMemListEnd;
	     ++ptrMemList)
	{
	    if (ptrMemList->mem == mem)
	    {
		memcpy(&ret, ptrMemList, sizeof(ret));
		--memListConsumed;
		if (ptrMemList != &memList[memListConsumed])
		{
		    memcpy(ptrMemList, 
			   &memList[memListConsumed], 
			   sizeof(*ptrMemList));
		}
		break;
	    }
	}
	
	if (ret.mem == lowAddress || ret.mem == highAddress) 
	{
	    findAddressExtremes();
	}
    }

    memtrackCheckPeriodic(AUTOCHECK_PERIOD);

    return ret;
}

static AllocRec* memList_find(void* mem)
{
    AllocRec* ptrMemList;
    AllocRec* ptrMemListEnd = &memList[memListConsumed];
    if (memList != NULL) 
    {
	for (ptrMemList = memList; 
	     ptrMemList < ptrMemListEnd;
	     ++ptrMemList)
	{
	    if (mem == ptrMemList->mem)
	    {
		return ptrMemList;
	    }
	}
    }
    return NULL;
}

static AllocRec* memList_findInRange(void* mem)
{
    AllocRec* ptrMemList;
    AllocRec* ptrMemListEnd = &memList[memListConsumed];
    if (memList != NULL) 
    {
	for (ptrMemList = memList; 
	     ptrMemList < ptrMemListEnd;
	     ++ptrMemList)
	{
	    if (mem >= ptrMemList->mem && 
		mem <= (void*)((char*)ptrMemList->mem + ptrMemList->size))
	    {
		return ptrMemList;
	    }
	}
    }
    return NULL;
}

static void memtrackCheckLeaks(void)
{
    extern unsigned long _ftext;
    extern unsigned long _fdata;

    static size_t lastLeakedSize = 0;

    int ix;
    size_t leakedSize;
    void** scanPoint;

    /* 
       Prevent recursive calls. This is necessary if this function
       calls something which eventualy calls a memtrack redefined function
       (like malloc). For example, Instead of fprintf, you may put in some
       logging function which behaves in this way. This is a safeguard
       against such future modifications.
    */
    static int in = 0;
    if (in) return; else in = 1;

    MUTEX_LOCK;

#if defined __alpha
    {
#       define REGISTER_COUNT 28
        void* registers[REGISTER_COUNT];

	/* Take snapshot of registers and check */
	registers[1] = (void*)asm("mov $1, %v0");
	registers[2] = (void*)asm("mov $2, %v0");
	registers[3] = (void*)asm("mov $3, %v0");
	registers[4] = (void*)asm("mov $4, %v0");
	registers[5] = (void*)asm("mov $5, %v0");
	registers[6] = (void*)asm("mov $6, %v0");
	registers[7] = (void*)asm("mov $7, %v0");
	registers[8] = (void*)asm("mov $8, %v0");
	registers[9] = (void*)asm("mov $9, %v0");
	registers[10] = (void*)asm("mov $10, %v0");
	registers[11] = (void*)asm("mov $11, %v0");
	registers[12] = (void*)asm("mov $12, %v0");
	registers[13] = (void*)asm("mov $13, %v0");
	registers[14] = (void*)asm("mov $14, %v0");
	registers[15] = (void*)asm("mov $15, %v0");
	registers[16] = (void*)asm("mov $16, %v0");
	registers[17] = (void*)asm("mov $17, %v0");
	registers[18] = (void*)asm("mov $18, %v0");
	registers[19] = (void*)asm("mov $19, %v0");
	registers[20] = (void*)asm("mov $20, %v0");
	registers[21] = (void*)asm("mov $21, %v0");
	registers[22] = (void*)asm("mov $22, %v0");
	registers[23] = (void*)asm("mov $23, %v0");
	registers[24] = (void*)asm("mov $24, %v0");
	registers[25] = (void*)asm("mov $25, %v0");
	registers[26] = (void*)asm("mov $26, %v0");
	registers[27] = (void*)asm("mov $27, %v0");

	for(ix = 1; ix < REGISTER_COUNT; ++ix)
	{
	    if (registers[ix] >= lowAddress && registers[ix] <= highAddress)
	    {
		AllocRec* mem;
		if ((mem = memList_find(registers[ix])) != NULL)
		{
		    mem->refExists |= REF_TO_START;
		}
		else if ((mem = memList_findInRange(registers[ix])) != NULL)
		{
		    mem->refExists |= REF_TO_WITHIN;
		}
	    }
	}
    }
#endif

    /* check stack */
    for (scanPoint = (void**)&scanPoint; scanPoint < (void**)&_ftext; ++scanPoint)
    {
	if (*scanPoint >= lowAddress && *scanPoint <= highAddress)
	{
	    AllocRec* mem;
	    if ((mem = memList_find(*scanPoint)) != NULL)
	    {
		mem->refExists |= REF_TO_START;
	    }
	    else if ((mem = memList_findInRange(*scanPoint)) != NULL)
	    {
		mem->refExists |= REF_TO_WITHIN;
	    }
	}
    } 

    /* check initialized data, BSS, and heap */
    SET_SIGSEGV_ACTION;

    for(scanPoint = (void**)&_fdata; scanPoint < (void**)sbrk(0); ++scanPoint)
    {
	if (SETJMP(jmpenv) == 0)
	{
	    for(; scanPoint < (void**)sbrk(0); ++scanPoint)
	    {
		if (scanPoint == (void**)&memList[0]) 
		{
		    scanPoint = (void**)&memList[memListAllocated];
		}
		else if (*scanPoint >= lowAddress && *scanPoint <= highAddress)
		{
		    AllocRec* mem;
		    if ((mem = memList_find(*scanPoint)) != NULL)
		    {
			mem->refExists |= REF_TO_START;
		    }
		    else if ((mem = memList_findInRange(*scanPoint)) != NULL)
		    {
			mem->refExists |= REF_TO_WITHIN;
		    }
		}
	    }
	}
    }

    RESET_SIGSEGV_ACTION;
    
    /* check environment */
    {
	extern char** environ;
	char** scanPoint;
	for (scanPoint = environ; *scanPoint; ++scanPoint)
	{
	    AllocRec* mem;
	    if ((mem = memList_find(*scanPoint)) != NULL)
	    {
		mem->refExists |= REF_TO_START;
	    }
	    else if ((mem = memList_findInRange(*scanPoint)) != NULL)
	    {
		mem->refExists |= REF_TO_WITHIN;
	    }
	}
    }
	
    leakedSize = 0;
    for (ix = 0; ix < memListConsumed; ++ix)
    {
	if ( memList[ix].refExists == NO_REF)
	{
	    if (memList[ix].reported != LEAK_REPORTED)
	    {
		fprintf(stderr,
			"memtrackCheckLeaks(): leaked %p size %d from %s:%d\n", 
			memList[ix].mem,
			memList[ix].size,
			memList[ix].fileName,
			memList[ix].fileLine);
		memList[ix].reported = LEAK_REPORTED;
	    }
	    leakedSize += memList[ix].size;
	} 
	else if ( (memList[ix].refExists & REF_TO_WITHIN) 
		  &&
		  ! (memList[ix].refExists & REF_TO_START) 
		  &&
		  memList[ix].reported != POTENTIAL_LEAK_REPORTED )
	{
	    fprintf(stderr,
		    "memtrackCheckLeaks(): potential leak %p size %d from %s:%d\n", 
		    memList[ix].mem,
		    memList[ix].size,
		    memList[ix].fileName,
		    memList[ix].fileLine);
	    memList[ix].reported = POTENTIAL_LEAK_REPORTED;
	}
	memList[ix].refExists = NO_REF;
    }

    if (leakedSize && leakedSize != lastLeakedSize)
    {
	fprintf(stderr,
		"memtrackCheckLeaks(): total leaked %d bytes\n", leakedSize);
    }
    lastLeakedSize = leakedSize;

    MUTEX_RELEASE;
    
    in = 0;
    return;
}

static void memtrackCheckSentinels(void)
{
    int ix;

    MUTEX_LOCK;
    
    for (ix = 0; ix < memListConsumed; ++ix)
    {
	char* mem = (char*)memList[ix].mem;
	if (mem[memList[ix].size-1] != END_SENTINEL)
	{
	    fprintf(stderr,
		    "memtrackCheckSentinels(): memory allocated at %s:%d is corrupted\n",
		    memList[ix].fileName,
		    memList[ix].fileLine);
	}
    }
    
    MUTEX_RELEASE;
    
    return;
}

/******************************************************************
 * The memtrackFree function will free the specified pointer unless its
 * clearly not a heap pointer.
 * If the pointer does not appear valid for any other reason, 
 * a message is sent to stderr.
 */
void memtrackFree(void* ptr, char* fn, ushort fl)
{
    AllocRec mem;

    if (ptr == NULL) 
    {
	fprintf(stderr, 
		"Error in memtrackFree: NULL being freed from %s:%d\n",
		fn, fl);
	return;
    }

    if (! isPtrHeap(ptr)) 
    {
	fprintf(stderr, 
		"Error in memtrackFree: ptr %p being freed from %s:%d is not on the heap\n", 
		ptr, fn, fl);
	return;
    }

    MUTEX_LOCK;
    
    mem = memList_remove(ptr);
    if (mem.mem == ptr)
    {
	markAsFree(mem.mem, mem.size);
    }
    else
    {
	AllocRec* range = memList_findInRange(ptr);
	if (range == NULL)
	{
	    /* 
	     * It is possible, and likely that the pointer is valid. 
	     * The pointer cannot be marked as free because the memory 
	     * tracker does not have valid information.
	     */
	    if (*(char*)ptr == FREED_FLAG)
	    {
		fprintf(stderr, 
			"Warning from memtrackFree: "
			"Possible multiple freeing of memory %p at %s:%d\n",
			ptr, fn, fl);
	    }
	    else
	    {
		fprintf(stderr, 
			"Warning from memtrackFree: "
			"ptr %p being freed at %s:%d is not known to the memory tracker\n",
			ptr, fn, fl);
	    }
	}
	else
	{
	    fprintf(stderr, 
		    "Warning from memtrackFree: "
		    "ptr %p freed at %s:%d does not point to the start of an allocation. "
		    "Perhaps you mean %p (size %d)\n",
		    ptr, fn, fl, range->mem, range->size);
	}

	/* Assume that at least one byte can be overwritten */
	markAsFree(ptr, 1);
    }

    MUTEX_RELEASE;

    free(ptr);

    return;
}

void* memtrackCalloc(size_t numElements, size_t elemSize, char* fn, ushort fl)
{
    volatile void* ret; /* Don't store in registers */

    ++numElements;
    ret = calloc(numElements, elemSize);
    ((char*)ret)[numElements * elemSize - 1] = END_SENTINEL;

    MUTEX_LOCK;
    memList_add(ret, numElements * elemSize, fn, fl);
    MUTEX_RELEASE;
    
    return (void*)ret;
}

void* memtrackMalloc(size_t size, char* fn, ushort fl)
{
    volatile void* ret;

    ++size;
    ret = malloc(size);
    memset((void*)ret, INIT_FLAG, size);
    ((char*)ret)[size - 1] = END_SENTINEL;

    MUTEX_LOCK;
    memList_add(ret, size, fn, fl);
    MUTEX_RELEASE;
    
    return (void*)ret;
}

void* memtrackValloc(size_t size, char* fn, ushort fl)
{
    volatile void* ret;

    ++size;
    ret = valloc(size);
    memset((void*)ret, INIT_FLAG, size);
    ((char*)ret)[size - 1] = END_SENTINEL;

    MUTEX_LOCK;
    memList_add(ret, size, fn, fl);
    MUTEX_RELEASE;
    
    return (void*)ret;
}

void* memtrackRealloc(void* ptr, size_t size, char* fn, ushort fl)
{
    volatile void* ret = NULL;

    if (size == 0) memtrackFree(ptr, fn, fl);
    else if (ptr == NULL) return memtrackMalloc(size, fn, fl);
    else if (! isPtrHeap(ptr)) 
    {
	fprintf(stderr, 
		"Error in memtrackRealloc: ptr %p at %s:%d is not on the heap\n", 
		ptr, fn, fl);
	return NULL;
    }
    else
    {
	AllocRec* mem;

	MUTEX_LOCK;

	mem = memList_find(ptr);

	if (mem == NULL)
	{
	    if ((mem = memList_findInRange(ptr)) != NULL)
	    {
		fprintf(stderr, 
			"Warning from memtrackRealloc: "
			"ptr %p at %s:%d does not point to the start of an allocation. "
			"Perhaps you mean %p (size %d)\n",
			ptr, fn, fl, mem->mem, mem->size);

		mem = NULL;
	    }
	    else if (*(char*)ptr == FREED_FLAG)
	    {
		fprintf(stderr, 
			"Warning from memtrackRealloc: "
			"Possible realloc of already freed memory %p at %s:%d\n", 
			ptr, fn, fl);
	    }
	    else
	    {
		fprintf(stderr, 
			"Warning from memtrackRealloc: "
			"ptr %p being realloced at %s:%d is not "
			"known to the memory tracker\n", 
			ptr, fn, fl);
	    }
	}

	++size;
	ret = realloc(ptr, size);
	((char*)ret)[size - 1] = END_SENTINEL;

	if (mem != NULL)
	{
	    mem->mem = (void*)ret;
	    mem->size = size;
	    findAddressExtremes();
	}
	else
	{
	    memList_add(ret, size, fn, fl);
	}

	MUTEX_RELEASE;
    }

    return (void*)ret;
}

char* memtrackStrdup(const char* s1, char* fn, ushort fl)
{	
    volatile char* ret;
    size_t size;

    if (s1 == NULL)
    {
	fprintf(stderr,
		"Error: memtrackStrdup called from %s:%d with NULL\n", fn, fl);
	return NULL;
    }

    size = strlen(s1) + 2;
    ret = malloc(size);
    strcpy((void*)ret, s1);
    ret[size - 1] = END_SENTINEL;

    MUTEX_LOCK;	
    memList_add((void*)ret, size, fn, fl);
    MUTEX_RELEASE;
    
    return (void*)ret;
}

char* memtrackStrcpy(char* dest, const char* src, size_t destSize, char* fn, ushort fl)
{
    long srcLen;
    int error = 0;

    if (dest == NULL || src == NULL)
    {
	fprintf(stderr,
		"Error: memtrackStrcpy called from %s:%d with bad params.\n",
		fn, fl);
	return NULL;
    }
    
    srcLen = strlen(src);

    if (destSize == sizeof(char*))
    {
	AllocRec* mem;

	MUTEX_LOCK;

	/* dest is probably with a pointer - check if the pointer is known to us */
	if ((mem = memList_find(dest)) != NULL)
	{
	    if (mem->size-1 <= srcLen)
	    {
		error = __LINE__;
		srcLen = mem->size-1-1;
	    }
	}
	else if ((mem = memList_findInRange(dest)) != NULL)
	{
	    size_t spaceLeft = mem->size-1 - (dest - (char*)mem->mem);
	    if (spaceLeft <= srcLen)
	    {
		error = __LINE__;
		srcLen = spaceLeft-1;
	    }
	}
	else if (*(char*)dest == FREED_FLAG)
	{
	    fprintf(stderr, 
		    "Warning from memtrackStrcpy: "
		    "Possible strcpy to already freed memory %p at %s:%d\n",
		    dest, fn, fl);
	}
	/* else dest is some pointer, and we cannot determine how much space it has */
	
	MUTEX_RELEASE;
    }
    else 
    {
	/* dest is probably an array of known size when this functions was called */
	if (destSize <= srcLen)
	{
	    error = __LINE__;
	    srcLen = destSize - 1;
	}
    }

    
    if (srcLen > 0)
    {
	strncpy(dest, src, srcLen);
	dest[srcLen] = 0;
    }
    else if (srcLen == 0)
    {
	*dest = 0;
    }

    if (error)
    {
	fprintf(stderr,
		"Error: memtrackStrcpy at %d called from %s:%d with too long src.\n",
		error, fn, fl);
    }
    
    return dest;
}

char* memtrackStrcat(char* dest, const char* src, size_t destSize, char* fn, ushort fl)
{
    long srcLen;
    size_t destLen;
    int error = 0;

    if (dest == NULL || src == NULL)
    {
	fprintf(stderr,
		"Error: memtrackStrcat called from %s:%d with bad params.\n",
		fn, fl);
	return NULL;
    }
    
    srcLen = strlen(src);
    destLen = strlen(dest);

    if (destSize == sizeof(char*))
    {
	AllocRec* mem;

	MUTEX_LOCK;

	/* dest is probably with a pointer - check if the pointer is known to us */
	if ((mem = memList_find(dest)) != NULL)
	{
	    if (mem->size-1 <= destLen + srcLen)
	    {
		error = __LINE__;
		srcLen = mem->size-1-destLen-1;
	    }
	}
	else if ((mem = memList_findInRange(dest)) != NULL)
	{
	    size_t spaceLeft = mem->size-1 - (dest - (char*)mem->mem);
	    if (spaceLeft <= destLen + srcLen)
	    {
		error = __LINE__;
		srcLen = spaceLeft-destLen-1;
	    }
	}
	else if (*(char*)dest == FREED_FLAG)
	{
	    fprintf(stderr, 
		    "Warning from memtrackStrcat: "
		    "Possible strcat to already freed memory %p at %s:%d\n",
		    dest, fn, fl);
	}
	/* else dest is some pointer, and we cannot determine how much space it has */
	
	MUTEX_RELEASE;
    }
    else 
    {
	/* dest is probably an array of known size when this functions was called */
	if (destSize <= destLen + srcLen)
	{
	    error = __LINE__;
	    srcLen = destSize - destLen - 1;
	}
    }

    if (srcLen > 0)
    {
	strncpy(dest + destLen, src, srcLen);
	dest[destLen + srcLen] = 0;
    }
    
    if (error)
    {
	fprintf(stderr,
		"Error: memtrackStrcat at %d called from %s:%d with too long src.\n",
		error, fn, fl);
    }
    
    return dest;
}

void memtrackCheckPeriodic(int period)
{
    static time_t lastLeakCheck = 0;

    if (time(NULL) - lastLeakCheck > period)
    {
	memtrackCheckLeaks();
	lastLeakCheck = time(NULL);
    }

    memtrackCheckSentinels();
	
    return;
}
